#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/resource.h>
#include <sys/file.h>
#include "mmlp.h"

#define TCP_PORT 9678

int init_socket(int offset, char* myname)
{
    struct hostent *hp;
    union {
        struct sockaddr    dummy;
        struct sockaddr_in in;
    } serv_addr;
    int sfd;

    if((hp = gethostbyname(myname)) == NULL) {
        fprintf(stderr, "%s: unknown host\n", myname);
        exit(1);
    }

    bzero((char *)&serv_addr, sizeof(serv_addr));
    serv_addr.in.sin_family = AF_INET;
    serv_addr.in.sin_port   = htons(TCP_PORT + offset);
    bcopy((char *)hp->h_addr, (char *)&serv_addr.in.sin_addr,
          hp->h_length);

    if((sfd = socket(AF_INET, SOCK_STREAM, NULL)) < 0)
    {
        perror("socket");
        exit(1);
    }

#ifdef SO_REUSEADDR
{int on=1; setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, (caddr_t)&on, sizeof(on));}
#endif /* SO_REUSEADDR */

    if(bind(sfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0)
    {
        perror("bind");
        close(sfd);
        exit(1);
    }

    if(listen(sfd, 5) < 0)
    {
        perror("listen");
        close(sfd);
        exit(1);
    }

    return sfd;
}

void wait_client(int sfd, int* fd, int n)
{
    struct hostent *hp;
    struct sockaddr client_data;
    int len;
    int i;

    len = sizeof(client_data);

    for(i = 0; i < n; i++)
    {
        if((fd[i] = accept(sfd, &client_data, &len)) < 0)
        {
            perror("accept");
            close(sfd);
            exit(1);
        }
	if((hp = gethostbyaddr((unsigned char *)&client_data.sa_data[2],
                               4, AF_INET)) == NULL)
            printf("?? is ready!\n");
        else
            printf("<%s> is ready!\n", hp->h_name);
    }
}

int connect_server(char *hostname, int offset)
{
    struct hostent *hp;
    union {
        struct sockaddr    dummy;
        struct sockaddr_in in;
    } serv_addr;
    int sfd;
    char* myname = get_env("HOST");

    if((hp = gethostbyname(hostname)) == NULL) {
        fprintf(stderr, "%s: unknown host\n", hostname);
        exit(1);
    }

    if((sfd = socket(AF_INET, SOCK_STREAM, NULL)) < 0)
    {
        perror("socket");
        exit(1);
    }

    bzero((char *)&serv_addr, sizeof(serv_addr));
    serv_addr.in.sin_family = AF_INET;
    serv_addr.in.sin_port   = htons(TCP_PORT + offset);
    bcopy((char *)hp->h_addr, (char *)&serv_addr.in.sin_addr, hp->h_length);

    if(connect(sfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) < 0){
        fprintf(stderr, "%s: %s: connection failed\n",
                myname, hostname);
        perror("connect");
        exit(1);
    }

    return sfd;
}
