#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include "mmlp.h"

#define MAX_CHILD 16

static void parse_remote_spec(char* name, char* h, char* p)
{
    char* d = name;

    while(*d && *d != ':')
    {
	*h++ = *d++;
    }

    if(!*d)
    {
	fprintf(stderr, "Invalid remote mml specified: `%s'\n", name);
	exit(1);
    }

    *h = '\0';
    d++;

    while(*d)
    {
	*p++ = *d++;
    }
    *p = '\0';
}

static void check_host_list(char** host_list, int nchild)
{
    char *new_host;
    int i;

    if(nchild <= 1)
	return;

    new_host = host_list[nchild - 1];

#ifndef AUDRIV
    for(i = 0; i < nchild - 1; i++)
    {
	if(!strcmp(host_list[i], new_host))
	{
	    fprintf(stderr, "Can't play the same host: `%s'\n", new_host);
	    exit(1);
	}
    }
#endif /* AUDRIV */
}

static void wait_all_children(int* pid_list, int nchild)
{
    int status;
    int pid;
    int i;
    int n;

    n = nchild;
    while(n)
    {
	pid = wait(&status);
	for(i = 0; i < nchild; i++)
	    if(pid == pid_list[i])
		n--;
    }
}

static void print_command(int argc, char** argv)
{
    int i;

    for(i = 0; i < argc; i++)
    {
	printf("%s ", argv[i]);
    }
    printf("\n");
}

static int execute_command(int argc, char** argv)
{
    int pid;

#ifdef DEBUG
    printf("remote exec: "), print_command(argc, argv);
#endif /* DEBUG */

    if((pid = fork()) < 0)
    {
	perror("fork");
	exit(1);
    }

    if(pid == 0)
    {
	execvp(argv[0], argv);
	perror(argv[0]);
	exit(1);
    }
    return pid;
}

static void sync_play_parent(int* fdlist, int nfd)
{
    int width, rval;
    fd_set readfds, origfds;
    int i;
    int n;
    char x;

    n = 0;
    FD_ZERO(&origfds);
    width = 0;
    for(i = 0; i < nfd; i++)
    {
	if(fdlist[i] >= 0)
	{
	    FD_SET(fdlist[i], &origfds);
	    if(fdlist[i] > width)
		width = fdlist[i];
	}
    }
    width++;

    while(1)
    {
	n = 0;
	readfds = origfds;
	do
	{
	    if((rval = select(width, &readfds, NULL, NULL, NULL)) < 0)
	    {
		perror("select");
		exit(1);
	    }

	    if(rval == 0)
		goto exit_parent;

	    for(i = 0; i < nfd; i++)
	    {
		if(FD_ISSET(fdlist[i], &readfds) != 0)
		{
		    if(read(fdlist[i], &x, 1) == 0)
			goto exit_parent;

		    FD_CLR(fdlist[i], &readfds);
		    n++;
		}
	    }
	} while(n < nfd);

	for(i = 0; i < nfd; i++)
	{
	    x = 1;
	    if(write(fdlist[i], &x, 1) == 0)
		goto exit_parent;
	    close(fdlist[i]);
	}

	return;
    }

  exit_parent:
    for(i = 0; i < nfd; i++)
	close(fdlist[i]);
    exit(0);
}


static int parent_fd = -1;
void sync_play(void)
{
    char x;

    if(parent_fd < 0)
	return;

    x = 0;
    write(parent_fd, &x, 1);
    if(read(parent_fd, &x, 1) == 0)
    {
	close(parent_fd);
	exit(0);
    }
}

void close_parent_fd(void)
{
    if(parent_fd >= 0)
    {
	close(parent_fd);
	parent_fd = -1;
    }
}

void sync_play_start(char* parent_host)
{
    parent_fd = connect_server(parent_host, port_offset);
    sync_play();
}

void exec_remote_clild(char** argv, int argc, int start_file_list)
{
    char** rsh_cmd_args;
    static char host_name[MAXLEN];
    static char mml_path[MAXLEN];
    static char buff[MAXLEN];
    char* host_name_list[MAX_CHILD];
    char* mml_path_list[MAX_CHILD];
    char* current_dir;
    char* current_host;
    int i;
    int nchild;
    int narg;
    int pid_list[MAX_CHILD];
    int sfd;
    int fdlist[MAX_CHILD];
    int mspart2 = 0;

    current_dir = get_env("PWD");
    current_host = get_env("HOST");

    nchild = 0;
    for(i = start_file_list; i < argc; i++)
    {
	if(nchild == MAX_CHILD)
	{
	    fprintf(stderr, "Too many surround mml data.\n");
	    exit(1);
	}

	parse_remote_spec(argv[i], host_name, buff);
	if(buff[0] != '/' &&
	   buff[0] != '~')
	{
	    sprintf(mml_path, "%s/%s", current_dir, buff);
	}
	else
	{
	    strcpy(mml_path, buff);
	}

	host_name_list[nchild] = strdup(host_name);
	mml_path_list[nchild]  = strdup(mml_path);
	if(host_name_list[nchild] == NULL ||
	   mml_path_list[nchild] == NULL)
	{
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	}
	nchild++;

	check_host_list(host_name_list, nchild);
    }

    rsh_cmd_args = (char **)malloc((argc + 6) * sizeof(char *));
    rsh_cmd_args[0] = "rsh";
    rsh_cmd_args[1] = NULL;	/* seting after */
    rsh_cmd_args[2] = argv[0];
    narg = 3;
    for(i = 1; i < start_file_list; i++)
    {
	if(!strcmp(argv[i], "-vol"))
	{
	    rsh_cmd_args[narg++] = "-vol";
	    rsh_cmd_args[narg++] = argv[++i];
	}
	else if(!strcmp(argv[i], "-Hz"))
	{
	    rsh_cmd_args[narg++] = "-Hz";
	    rsh_cmd_args[narg++] = argv[++i];
	}
	else if(!strcmp(argv[i], "-port"))
	{
	    rsh_cmd_args[narg++] = "-port";
	    rsh_cmd_args[narg++] = argv[++i];
	}
	else if(!strcmp(argv[i], "-ms"))
	{
	    rsh_cmd_args[narg++] = "-mspart";
	    mspart2 = narg;
	    rsh_cmd_args[narg++] = "0";
	    mono_sound_size = 2;
	}
	else if(!strcmp(argv[i], "-nms"))
	{
	    rsh_cmd_args[narg++] = "-mspart";
	    mono_sound_size = atoi(argv[++i]);
	    mspart2 = narg;
	    rsh_cmd_args[narg++] = "0";
	}
    }

    rsh_cmd_args[narg++] = "-surround-child";
    rsh_cmd_args[narg++] = current_host;
    rsh_cmd_args[narg++] = NULL; /* seting after */
    rsh_cmd_args[narg] = NULL; /* terminating command */

    sfd = init_socket(port_offset, current_host);

    for(i = 0; i < nchild; i++)
    {
	if(mspart2)
	{
	    sprintf(buff, "%d/%d", i % mono_sound_size + 1, mono_sound_size);
	    rsh_cmd_args[mspart2] = buff;
	}

	rsh_cmd_args[1] = host_name_list[i];
	rsh_cmd_args[narg - 1] = mml_path_list[i];
	pid_list[i] = execute_command(narg, rsh_cmd_args);
    }

    wait_client(sfd, fdlist, nchild);
    close(sfd);
    sync_play_parent(fdlist, nchild);

/*    printf("## wait exit childred\n"); */

    wait_all_children(pid_list, nchild);
    exit(0);
}
