#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mmlp.h"

#define DEFAULT_STACK_SIZE 16

/* textbuff
 * for imprementation follows command:
 * |: ... :|   repeat the phase enclosed with |: and :|
 * |: n ... :| repeat n times the phase enclosed with |: and :|
 */

struct textbuff
{
    char* text;
    struct textbuff* next;	/* next text */
};

/*
 * All text must be stored to repeat the command
+----+  +----+
|text|->|text|-> NULL
+----+  +----+
*/

struct textbuffStack
{
    int loop_counter;
    int lineno;
    int channel;
    char* text;
    struct textbuff* pos;
};
/* Store the current text status */

static struct textbuff* current_pos;
static struct textbuff* last_pos;
static struct textbuffStack* textbuffStack;
static int textbuffStack_ptr;
static int textbuffStack_size = DEFAULT_STACK_SIZE;
static struct textbuff* first_text_buff;

void print_all_text_buff(void)
{
    struct textbuff* ptr = first_text_buff;

    fprintf(stderr, "## print-all-text-buff\n");
    while(ptr)
    {
	fprintf(stderr, "## <=%s", ptr->text);
	ptr = ptr->next;
    }
}

void add_mmltext(char* text)
{
    struct textbuff* node;

    node = (struct textbuff *)xmalloc(sizeof(struct textbuff));
    node->text = text;
    node->next = NULL;

    if(first_text_buff == NULL)
    {
	first_text_buff = current_pos = last_pos = node;
    }
    else
    {
	last_pos->next = node;
	last_pos = node;
	if(current_pos == NULL)
	    current_pos = node;
    }
}

char* get_next_mmltext(void)
{
    if(current_pos == NULL)
	return NULL;

    current_pos = current_pos->next;

    if(current_pos == NULL)
	return NULL;
    return current_pos->text;
}


void begin_loop(int n)
{
    if(textbuffStack == NULL)
    {
	textbuffStack = (struct textbuffStack *)
	    xmalloc(textbuffStack_size * sizeof(struct textbuffStack));
    }
    else if(textbuffStack_size == textbuffStack_ptr)
    {
	textbuffStack_size *= 2;
	textbuffStack = (struct textbuffStack *)
	    realloc(textbuffStack,
		    textbuffStack_size * sizeof(struct textbuffStack));
	if(textbuffStack == NULL)
	{
	    fprintf(stderr, "Memory exhausted.\n");
	    quit(1);
	}
    }

    textbuffStack[textbuffStack_ptr].loop_counter = n;
    textbuffStack[textbuffStack_ptr].lineno = lineno;
    textbuffStack[textbuffStack_ptr].channel = current_channel;
    textbuffStack[textbuffStack_ptr].text = mml_text;
    textbuffStack[textbuffStack_ptr].pos = current_pos;
    textbuffStack_ptr++;
}

void end_loop(void)
{
    int x = textbuffStack_ptr - 1;

    if(x < 0)
    {
	if(!is_beginning_of_line)
	    fputc('\n', stderr);
	fprintf(stderr, "illegal loop command, in %d\n", lineno);
	quit(1);
    }

    if(textbuffStack[x].loop_counter <= 0)
    {
	textbuffStack_ptr = x;
	return;
    }

    if(trace_mode)
	fprintf(stderr, "|:%d", textbuffStack[x].loop_counter);

    textbuffStack[x].loop_counter--;
    lineno          = textbuffStack[x].lineno;
    current_channel = textbuffStack[x].channel;
    mml_text        = textbuffStack[x].text;
    current_pos     = textbuffStack[x].pos;
}
