#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "mmlp.h"

int token_type;
int token_val;
char* mml_text;
int lineno = 1;
int trace_mode = 1;
int is_beginning_of_line = 1;
static char token_buff[MAXLEN];

static void skip_line(void)
{
    while(*mml_text && *mml_text != '\n')
    {
	if(*mml_text == '\\' && *(mml_text + 1) == '\n')
	{
	    mml_text += 2;
	    lineno++;
	}
	else
	    mml_text++;

    }
    if(*mml_text == '\n')
    {
	mml_text++;
	lineno++;
    }
    is_beginning_of_line = 1;
}

static void skip_space(void)
{
    do
    {
	if(is_beginning_of_line)
	{
	    while(*mml_text == '\\')
	    {
		ext_command(mml_text + 1);
		skip_line();
	    }
	}

	while(*mml_text == ' ' ||
	      *mml_text == '\t')
	    mml_text++;

	if(*mml_text == '\\' && *(mml_text + 1) == '\n')
	{
	    mml_text += 2;
	    lineno++;
	    continue;
	}
	if(*mml_text == '%')
	{
	    if(trace_mode)
		fprintf(stderr, "%d: %%", lineno);

	    while(*mml_text != '\n' && *mml_text != '\0')
	    {
		mml_text++;
		if(trace_mode)
		    fputc(*mml_text, stderr);
	    }

	    is_beginning_of_line = 1;

	    if(*mml_text == '\n')
	    {
		mml_text++;
		lineno++;
		continue;
	    }
	    else if(trace_mode)
		fputc('\n', stderr);

	}

	if(*mml_text == '\0')
	{
	    char* next_text;
	    next_text = get_next_mmltext();
	    if(next_text == NULL)
		break;
	    mml_text = next_text;
	    continue;
	}
	break;
    } while(1);
}

int gettoken(void)
{
    register char* tmp;
    static int offset_level[] = {9, 11, 0, 2, 4, 5, 7};
			      /* A  B   C  D  E  F  G */
    static int diff_code_level[12] = {0,0,0,0,0,0,0,0,0,0,0,0};

  next_get_token:
    skip_space();

    if(*mml_text == '\0')
    {
	char* next_text;

	next_text = get_next_mmltext();
	if(next_text == NULL)
	{
	    token_type = EOF;
	    token_val = *mml_text;
	    return 1;
	}
	mml_text = next_text;
	goto next_get_token;
    }

    if(trace_mode)
    {
	if(is_beginning_of_line)
	    fprintf(stderr, "%d: ", lineno);
	fputc(*mml_text, stderr);
	is_beginning_of_line = 0;
    }

    switch(*mml_text)
    {
      case '~':
	mml_text++;
	skip_space();
	fputc(*mml_text, stderr);
	if(*mml_text < 'A' || *mml_text > 'G')
	{
	    illegal_command();
	    goto next_get_token;
	}
	token_val = offset_level[*mml_text++ - 'A'];
	skip_space();
	fputc(*mml_text, stderr);
	if(*mml_text == '+' || *mml_text == '#')
	    diff_code_level[token_val] = 1;
	else if(*mml_text == '-')
	    diff_code_level[token_val] = -1;
	else if(*mml_text == '/')
	    diff_code_level[token_val] = 0;
	else
	{
	    illegal_command();
	    goto next_get_token;
	}
	mml_text++;
	goto next_get_token;

      case 'A':
      case 'B':
      case 'C':
      case 'D':
      case 'E':
      case 'F':
      case 'G':
	token_type = PLAY_SCALE;
	token_val = offset_level[*mml_text++ - 'A'];
	skip_space();
	if(*mml_text == '+' || *mml_text == '#')
	{
	    if(trace_mode)
		fputc(*mml_text, stderr);
	    token_val = (1 + token_val) % 12;
	    mml_text++;
	}
	else if(*mml_text == '-')
	{
	    if(trace_mode)
		fputc('-', stderr);
	    token_val = (11 + token_val) % 12;
	    mml_text++;
	}
	else if(*mml_text == '/')
	{
	    if(trace_mode)
		fputc('/', stderr);
	    mml_text++;
	}
	else
	{
	    token_val = (token_val + 12 + diff_code_level[token_val]) % 12;
	}
	return 1;
      case 'N':
	token_type = PLAY_SCALEN;
	mml_text++;
	if(!isdigit(*mml_text))
	{
	    illegal_command();
	    goto next_get_token;
	}
	tmp = token_buff;
	if(trace_mode)
	{
	    *tmp++ = *mml_text++;
	    while(isdigit(*mml_text))
	    {
		fputc(*mml_text, stderr);
		*tmp++ = *mml_text++;
	    }
	}
	else
	{
	    while(isdigit(*mml_text))
		*tmp++ = *mml_text++;
	}
	*tmp++ = '\0';
	token_val = atoi(token_buff);
	return 1;
      case '+':
      case '#':
      case '-':
      case '.':
      case '&':
	token_type = OP;
	token_val = *mml_text++;
	return 1;
      case 'L':
      case 'Q':
      case 'O':
      case 'T':
      case 'R':
      case 'M':
      case 'S':
      case '<':
      case '>':
      case '{':
      case '}':
	token_type = COMMAND;
	token_val = *mml_text++;
	return 1;
      case 'V':
	token_type = COMMAND;
	token_val = 'V';
	mml_text++;
	return 1;
      case '@':
	token_type = COMMAND;
	if(*(mml_text + 1) == 'V')
	{
	    if(trace_mode)
		fputc('V', stderr);
	    token_val = 'v';
	    mml_text += 2;
	    return 1;
	}
	token_val = *mml_text++;
	return 1;
      case 'Y':
	token_type = COMMAND;
	token_val = *mml_text++;
	return 1;
      case ',':
	token_type = EOPart;
	token_val = *mml_text++;
	return 1;
      case '\n':
      case '\r':
	token_type = EOPhase;
	token_val = *mml_text++;
	lineno++;
	is_beginning_of_line = 1;
	return 1;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	tmp = token_buff;
	if(trace_mode)
	{
	    *tmp++ = *mml_text++;
	    while(isdigit(*mml_text))
	    {
		fputc(*mml_text, stderr);
		*tmp++ = *mml_text++;
	    }
	}
	else
	{
	    while(isdigit(*mml_text))
		*tmp++ = *mml_text++;
	}
	*tmp++ = '\0';
	token_type = NUMBER;
	token_val = atoi(token_buff);
	return 1;
      case '|':
	token_type = START_LOOP;
	mml_text++;
	skip_space();
	if(*mml_text != ':')
	{
	    illegal_command();
	    goto next_get_token;
	}
	if(trace_mode)
	    fputc(':', stderr);
	mml_text++;
	skip_space();

	if(isdigit(*mml_text))
	{
	    tmp = token_buff;
	    if(trace_mode)
	    {
		fputc(*mml_text, stderr);
		*tmp++ = *mml_text++;
		while(isdigit(*mml_text))
		{
		    fputc(*mml_text, stderr);
		    *tmp++ = *mml_text++;
		}
	    }
	    else
	    {
		while(isdigit(*mml_text))
		    *tmp++ = *mml_text++;
	    }
	    *tmp++ = '\0';
	    token_val = atoi(token_buff) - 1;
	}
	else
	{
	    token_val = 1;
	    if(trace_mode)
		fputc('2', stderr);
	}
	return 1;
      case ':':
	token_type = END_LOOP;
	mml_text++;
	skip_space();
	if(*mml_text != '|')
	{
	    illegal_command();
	    goto next_get_token;
	}
	mml_text++;
	if(trace_mode)
	    fputc('|', stderr);
	return 1;
    }

    if(trace_mode && !is_beginning_of_line)
	fputc('\n', stderr);
    fprintf(stderr, "Unknown character `%c' (0x%02x) in %d\n",
	    *mml_text, *mml_text, lineno);

    mml_text++;
    goto next_get_token;
}
